# frozen_string_literal: true
module Hermes
  module Api
    module Dmo
      # Base Controller for Hermes app (Parent Class)
      class ApplicationController < ::Api::D2::ApplicationController
        # protect_from_forgery with: :exception
        @@engine_path = Hermes::Engine.root
        require_sugar_tree = %w[lib hermes tools require_sugar]

        require @@engine_path.join(*require_sugar_tree).to_s

        serialization_scope :current_user
        respond_to :json

        @@base_set_instance_for = %i[show update destroy]

        before_action :allow_true_user, :check_if_log_as, if: :true_user?
        after_filter :log_action

        def log_action
          status = response.code
          p = params.except(:controller, :action, :uuid, :doc, :space, :auth_token, :file)
          Log.create(user_id: (current_user && current_user.id), code: "API-#{controller_path}", action: "status : #{status} | params : #{p}")
        end

        def initialize
          @query_precision = {}
          @options = {}
          @module_name = 'Hermes'
          super
          init_params_and_operators
        end

        @@mutate_params_only = %i[readers writers]
        class << self
          def set_instance_for(*arr)
            @set_instance_for ||= @@base_set_instance_for
            @set_instance_for = @set_instance_for.concat(arr).uniq.compact
          end

          def mutate_params_only(*array)
            @@mutate_params_only = @@mutate_params_only.concat(array)
                                                       .uniq
                                                       .compact
          end
        end

        def init_params_and_operators
          @additional_params = {}
          @acceptable_params = [:created_at, :updated_at]
          @regex_default_params =
            @regex_params =
              @valid_operators = []
        end

        before_action :set_instance, only: @set_instance_for, unless: :skip_set_instance?

        def handle_order_param(val)
          return unless [1, -1, '1', '-1', 'asc', 'desc', :asc, :desc].include?(val)

          val = Integer(val) if ['1', '-1'].include?(val)
          if [1,-1].include? val
            val == 1 ? :asc : :desc
          elsif %w[asc desc].include?(val)
            val.to_sym
          else
            val
          end
        end

        def order_params
          if %w[
            ActionController::Parameters
            Mongoid::Document
            Hash
          ].include?(params[:sort].class.name)
            params[:sort].each_with_object({}) do |(key, val), h|
              next unless handle_order_param(val)

              h[key] = handle_order_param(val)
            end
          end
        end

        def include_relationships
          []
        end

        def index_aggregation(target_model: @model, no_acl: false)
          return unless target_model

          precision = @query_precision || {}
          precision[:user_id] = params[:responsible] if params[:responsible]
          query = target_model
          query = query.readable_by(@user) unless no_acl
          query.where(build_query(target_model).deep_merge(precision))
               .order_by(order_params)
               .includes(include_relationships)
        end

        # a utilser pour les parametres d'update et create
        def mutate_valid_params
          ActiveModelSerializers::Deserialization.jsonapi_parse!(
            params,
            only: @@mutate_params_only, key_transform: :unaltered
          ).merge(@additional_params)
        end

        def index
          paginate(
            index_aggregation,
            each_serializer: @serializer,
            **@options
          )
        end

        def show
          return render_json_error(:not_found, :object_not_found) unless @instance

          render(
            jsonapi: @instance,
            serializer: @serializer,
            **@options
          )
        end

        def create
          create_without_render
          return respond_with_errors(@instance) unless @instance.save

          render(jsonapi: @instance, serializer: @serializer, status: :created)
        end

        def create_without_render
          user_id = @user.id
          param_user = params&.fetch(:data, nil)&.fetch(:attributes, nil)&.fetch(:user_id, nil)
          Rails.logger.debug "TEST before #{user_id}"
          user_id = param_user if @user.has_role?('Administrator') && param_user
          Rails.logger.info "TEST #{mutate_valid_params} after #{user_id}"

          @instance = @model.new(mutate_valid_params.merge(user_id: user_id))
        end

        def update
          return render_json_error(:not_found, :object_not_found) unless @instance
          return render_json_error(:unprocessable_entity, :not_writers) unless @instance.reload && @instance.writable_by?(@user)
          check_params_and_update
        end

        def check_params_and_update
          valid_params = mutate_valid_params
          param_user = params&.fetch(:data, nil)&.fetch(:attributes, nil)&.fetch(:user_id, nil)
          valid_params[:user_id] = param_user if param_user && @user.has_role?('Administrator')
          return render_json_error(:parameters_not_present, :no_valid_parameters) unless valid_params.present?
          Rails.logger.info "TEST valid_params are #{valid_params}"
          return respond_with_errors(@instance) unless @instance.update(valid_params) && @instance.save

          render(jsonapi: @instance, serializer: @serializer)
        end

        def reservataire_not_matching_id
          current_user_roles.include?('reservataire') && @user.reservation_holder_id != @instance.reservation_holder_id
        end

        def user_can_destroy?
         is_writer_and_is_owner = (@instance.user_id.to_s == current_user.id.to_s) &&
             (current_user.has_role?('Administrator') ||
                 current_user.has_role?('Manager'))
          is_writer_and_is_owner || current_user.has_role?('Superadmin')
        end

        def destroy
          return render_json_error(:not_found, :instance_not_found) unless @instance
          return render_json_error(:not_found, :not_admin_or_manager) unless user_can_destroy?
          return render(json: @instance.errors, status: :unprocessable_entity) unless @instance.destroy

          render(body: nil, status: :no_content)
        end

        def set_instance
          @instance = @model.readable_by(@user).find_by(id: params[:id], **@query_precision) if params[:id]
        end

        def allow_true_user
          allowed = ['superadministrateur', 'collaborateur offre logement'] & @true_user.attributes.slice(:roles, :auth_roles).values.flatten.uniq.compact
          @allow_true_user ||= allowed.any?
        end

        def current_user_roles
          current_user.attributes.slice(:roles, :auth_roles).values.flatten.uniq.compact
        end

        def current_user
          @user
        end

        def current_lessor
          (current_user_roles & ['bailleur']).present? ? current_user.lessor : nil
        end

        def per_page
          @per_apge ||= params[:per_page]&.to_i || 15
        end

        def page
          @page ||= (param_page = params[:page]).present? ? (p = param_page.to_i) <= 0 ? 1 : p : 1
        end

        def flatten_hash(param, prefix=nil)
          param.each_pair.reduce({}) do |a, (k, v)|
            v.is_a?(Hash) ? a.merge(flatten_hash(v, "#{prefix}#{k}.")) : a.merge("#{prefix}#{k}".to_sym => v)
          end
        end

        def paginate_like_array(collection, mapped_attributes = {})

          count = collection.count
          total_pages = (t_p = (count.to_f / per_page).ceil) >= 1 ? t_p : 1

          self_link = request.original_url.to_page(page)
          first_link = self_link.to_page(1)
          prev_link = (prev_p = page - 1).positive? ? self_link.to_page(prev_p) : nil
          next_link = (next_p = page + 1) > total_pages ? nil : self_link.to_page(next_p)
          last_link = self_link.to_page(total_pages)

          paginated_entries = collection.paginate(per_page: per_page, page: page)

          if mapped_attributes.length
            paginated_entries = paginated_entries.map do |e|
              data = {}
              mapped_attributes.each_key do |key|
                data[key] = eval("e#{mapped_attributes[key]}")
              end
              data
            end
          end

          render jsonapi: {
            data: paginated_entries,
            links: {
              self: self_link,
              first: first_link,
              prev: prev_link,
              next: next_link,
              last: last_link
            },
            meta: {
              pagination: {
                per_page: per_page,
                total_pages: total_pages,
                total_objects: count,
                links: {
                  first: first_link[/\/api\/.*/],
                  last: last_link[/\/api\/.*/]
                }
              }
            }
          }
        end

        def paginate(collection, **options)
          if collection.is_a? Mongo::Collection::View::Aggregation
            identifier = options[:entries_identifier] || '_id'
            entries_class = options[:entries_class] || @model
            serializer = options[:each_serializer] || @serializer
            field_identifier = options[:field_identifier] || 'id'

            count = collection.count
            total_pages = (t_p = (count.to_f / per_page).ceil) >= 1 ? t_p : 1

            self_link = request.original_url.to_page(page)
            first_link = self_link.to_page(1)
            prev_link = (prev_p = page - 1).positive? ? self_link.to_page(prev_p) : nil
            next_link = (next_p = page + 1) > total_pages ? nil : self_link.to_page(next_p)
            last_link = self_link.to_page(total_pages)

            paginated_entries = collection.paginate(per_page: per_page, page: page)
            objects = paginated_entries.map do |entry|
              next unless entry

              object_id = identifier.split(".")
                                    .inject(entry) { |hash, key| hash&.fetch(key, nil) }
              next unless object_id

              entries_class.find_by(field_identifier => object_id)
            end
            fake_serial = objects.map do |object|
              next unless object

              serializer.new(
                object,
                scope: current_user,
                root: true,
                **options
              )
            end

            render jsonapi: {
              data: fake_serial.map do |entry|
                next unless entry

                {
                  id: entry.id,
                  type: entry._type,
                  attributes: entry.attributes.reject { |key| ['id', :id].include? key }
                }
              end,
              links: {
                self: self_link,
                first: first_link,
                prev: prev_link,
                next: next_link,
                last: last_link
              },
              meta: {
                pagination: {
                  per_page: per_page,
                  total_pages: total_pages,
                  total_objects: count,
                  links: {
                    first: first_link[/\/api\/.*/],
                    last: last_link[/\/api\/.*/]
                  }
                }
              }
            }
          else
            super(collection, **options)
          end
        end

        protected

        def true_user?
          @true_user.present?
        end

        def check_if_log_as
          unless @allow_true_user
            render json: {
              success: false,
              message: 'Action not allowed'
            }, status: :unauthorized
          end
        end

        def skip_set_instance?
          ['notifications', 'subscriptions'].include? controller_name
        end
      end
    end
  end
end
